#
# Copyright 2022 Ocean Protocol Foundation
# SPDX-License-Identifier: Apache-2.0
#
import importlib
import json
import logging
import os
from pathlib import Path
from typing import Any, Dict, Optional, Union

from jsonsempai import magic  # noqa: F401
from addresses import address as contract_addresses  # noqa: F401
from enforce_typing import enforce_types
from web3 import WebsocketProvider
from web3.contract import Contract
from web3.exceptions import ExtraDataLengthError
from web3.main import Web3
from web3.middleware import geth_poa_middleware

import artifacts  # noqa
from ocean_lib.web3_internal.web3_overrides.http_provider import CustomHTTPProvider

logger = logging.getLogger(__name__)
GANACHE_URL = "http://127.0.0.1:8545"


@enforce_types
def get_contract_definition(contract_name: str) -> Dict[str, Any]:
    """Returns the abi JSON for a contract name."""
    try:
        return importlib.import_module("artifacts." + contract_name).__dict__
    except ModuleNotFoundError:
        raise TypeError("Contract name does not exist in artifacts.")


@enforce_types
def load_contract(web3: Web3, contract_name: str, address: Optional[str]) -> Contract:
    """Loads a contract using its name and address."""
    contract_definition = get_contract_definition(contract_name)
    abi = contract_definition["abi"]
    bytecode = contract_definition["bytecode"]
    contract = web3.eth.contract(address=address, abi=abi, bytecode=bytecode)
    return contract


@enforce_types
def get_addresses_with_fallback(config):
    address_file = config.get("ADDRESS_FILE")
    address_file = (
        os.path.expanduser(address_file)
        if address_file
        else Path(contract_addresses.__file__).expanduser().resolve()
    )

    if not address_file or not os.path.exists(address_file):
        raise Exception("Address file not found.")
    with open(address_file) as f:
        addresses = json.load(f)

    return addresses


@enforce_types
def get_contracts_addresses(config) -> Optional[Dict[str, str]]:
    """Get addresses for all contract names, per network and address_file given."""
    if "CHAIN_ID" not in config:
        w3 = get_web3(config["RPC_URL"])
        # cache it to prevent further calls
        config["CHAIN_ID"] = w3.eth.chain_id

    chain_id = config["CHAIN_ID"]
    addresses = get_addresses_with_fallback(config)

    network_addresses = [
        val for key, val in addresses.items() if val["chainId"] == chain_id
    ]

    if network_addresses is None:
        raise Exception(
            f"Address not found for {chain_id}. Please check your address file."
        )

    return _checksum_contract_addresses(network_addresses=network_addresses[0])


@enforce_types
# Check singnet/snet-cli#142 (comment). You need to provide a lowercase address then call web3.toChecksumAddress()
# for software safety.
def _checksum_contract_addresses(
    network_addresses: Dict[str, Any]
) -> Optional[Dict[str, Any]]:
    for key, value in network_addresses.items():
        if key == "chainId":
            continue
        if isinstance(value, int):
            continue
        if isinstance(value, dict):
            for k, v in value.items():
                value.update({k: Web3.toChecksumAddress(v.lower())})
        else:
            network_addresses.update({key: Web3.toChecksumAddress(value.lower())})

    return network_addresses


@enforce_types
def get_web3(network_url: Optional[str] = None) -> Web3:
    """
    Return a web3 instance connected via the given network_url.
    Adds POA middleware if needed.
    """
    if not network_url:
        network_url = GANACHE_URL

    provider = get_web3_connection_provider(network_url)
    web3 = Web3(provider)

    try:
        web3.eth.getBlock("latest")
    except ExtraDataLengthError:
        web3.middleware_onion.inject(geth_poa_middleware, layer=0)

    return web3


@enforce_types
def get_web3_connection_provider(
    network_url: str,
) -> Union[CustomHTTPProvider, WebsocketProvider]:
    """Return the suitable web3 provider based on the network_url.

    Requires going through some gateway such as `infura`.

    Using infura has some issues if your code is relying on evm events.
    To use events with an infura connection you have to use the websocket interface.

    Make sure the `infura` url for websocket connection has the following format
    wss://goerli.infura.io/ws/v3/357f2fe737db4304bd2f7285c5602d0d
    Note the `/ws/` in the middle and the `wss` protocol in the beginning.

    :param network_url: str
    :return: provider : Union[CustomHTTPProvider, WebsocketProvider]
    """
    if network_url.startswith("http"):
        return CustomHTTPProvider(network_url)
    elif network_url.startswith("ws"):
        return WebsocketProvider(network_url)
    else:
        msg = (
            f"The given network_url *{network_url}* does not start with either"
            f"`http` or `wss`. A correct network url is required."
        )
        raise AssertionError(msg)
